from warnings import warn

from django.conf import settings
from django.contrib.auth.models import Group
from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.color import color_style
from django.urls.base import reverse
from edc_auth.models import Role

style = color_style()


def get_or_create_group(group_name):
    try:
        group = Group.objects.get(name=group_name)
    except ObjectDoesNotExist:
        group = Group.objects.create(name=group_name)
    return group


def login(
    testcase,
    user=None,
    superuser=None,
    groups=None,
    roles=None,
    sites=None,
    redirect_url=None,
):
    user = testcase.user if user is None else user
    user.is_superuser = True if superuser is None else superuser
    if user.is_superuser:
        warn(
            style.WARNING(
                "\nWarning: Running test where logged in user is a superuser. See login() \n"
            )
        )
    user.is_active = True
    user.is_staff = True
    if not user.is_superuser:
        if groups:
            user.groups.clear()
        for group_name in groups or []:
            group = Group.objects.get(name=group_name)
            user.groups.add(group)
        if roles:
            user.userprofile.roles.clear()
        for role_name in roles or []:
            role = Role.objects.get(name=role_name)
            user.userprofile.roles.add(role)
        for site in Site.objects.filter(id__in=sites or [settings.SITE_ID]):
            user.userprofile.sites.add(site)
    user.userprofile.save()
    user.save()
    user.refresh_from_db()
    form = (
        testcase.app.get(reverse(redirect_url or settings.LOGIN_REDIRECT_URL))
        .maybe_follow()
        .form
    )
    form["username"] = user.username
    form["password"] = "pass"  # nosec B105
    return form.submit()



import sys

import django
from django.conf import settings
from django.test.runner import DiscoverRunner


def func_main(app_name, project_settings):
    if not settings.configured:
        settings.configure(**project_settings)
    django.setup()
    tags = [t.split("=")[1] for t in sys.argv if t.startswith("--tag")]
    failfast = any([True for t in sys.argv if t.startswith("--failfast")])
    failures = DiscoverRunner(failfast=failfast, tags=tags).run_tests([f"{app_name}.tests"])
    sys.exit(failures)



import os
from unittest import mock
from unittest.case import TestCase

from ..default_test_settings import DefaultTestSettings


class TestUtils(TestCase):
    def test_(self):

        os.environ["TRAVIS"] = ""
        base_dir = os.path.dirname(os.path.abspath(__file__))

        with mock.patch("sys.argv", ["tests.py"]):
            DEFAULT_SETTINGS = DefaultTestSettings(
                calling_file=__file__,
                BASE_DIR=base_dir,
                APP_NAME="edc_test_utils",
                ETC_DIR=os.path.join(base_dir, "etc"),
            ).settings

        self.assertIn("sqlite", DEFAULT_SETTINGS.get("DATABASES").get("default").get("ENGINE"))

    def test_3(self):
        base_dir = os.path.dirname(os.path.abspath(__file__))
        os.environ["TRAVIS"] = "True"
        with mock.patch("sys.argv", ["tests.py"]):
            DEFAULT_SETTINGS = DefaultTestSettings(
                calling_file="tests.py",
                BASE_DIR=base_dir,
                APP_NAME="edc_test_utils",
                ETC_DIR=os.path.join(base_dir, "etc"),
            ).settings

        self.assertIn("mysql", DEFAULT_SETTINGS.get("DATABASES").get("default").get("ENGINE"))

    def test_encryption_keys(self):
        with mock.patch("sys.argv", ["tests.py"]):
            base_dir = os.path.dirname(os.path.abspath(__file__))
            DEFAULT_SETTINGS = DefaultTestSettings(
                calling_file="tests.py",
                app_name="edc_test_utils",
                base_dir=base_dir,
                installed_apps=["django_crypto_fields.apps.AppConfig"],
                etc_dir=os.path.join(base_dir, "etc"),
            ).settings

        self.assertTrue(os.path.exists(DEFAULT_SETTINGS.get("KEY_PATH")))
        self.assertIn("AUTO_CREATE_KEYS", DEFAULT_SETTINGS)



from unittest.case import TestCase

from django.apps import apps as django_apps
from edc_list_data.model_mixins import ListModelMixin


class NaturalKeyTestHelperError(Exception):
    pass


class NaturalKeyTestHelper(TestCase):
    def nk_test_natural_key_attr(self, *app_labels, exclude_models=None):

        """Asserts all models in given apps have a
        natural_key model method.
        """
        exclude_models = exclude_models or []
        for app_label in app_labels:
            models = django_apps.get_app_config(app_label).get_models()
            for model in models:
                if (
                    model._meta.label_lower not in exclude_models
                    and not issubclass(model, ListModelMixin)
                    and ".tests." not in model.__module__
                ):
                    self.assertTrue(
                        "natural_key" in dir(model),
                        "Model method 'natural_key' missing. "
                        f"Got '{model._meta.label_lower}'.",
                    )

    def nk_test_get_by_natural_key_attr(self, *app_labels, exclude_models=None):
        """Asserts all models in given apps have a get_by_natural_key
        manager method.
        """
        exclude_models = exclude_models or []
        for app_label in app_labels:
            models = django_apps.get_app_config(app_label).get_models()
            for model in models:
                if (
                    model._meta.label_lower not in exclude_models
                    and not issubclass(model, ListModelMixin)
                    and ".tests." not in model.__module__
                ):
                    self.assertTrue(
                        "get_by_natural_key" in dir(model.objects),
                        "Manager method 'get_by_natural_key' missing. "
                        f"Got '{model._meta.label_lower}'.",
                    )

    def nk_test_natural_keys(self, complete_required_crfs):
        """Asserts tuple from natural_key when passed to
        get_by_natural_key successfully gets the model instance.
        """
        for objs in complete_required_crfs.values():
            for obj in objs:
                options = obj.natural_key()
                try:
                    obj.__class__.objects.get_by_natural_key(*options)
                except obj.__class__.DoesNotExist:
                    self.fail(
                        "get_by_natural_key query failed for "
                        f"'{obj._meta.label_lower}' with "
                        f"options {options}."
                    )
                except TypeError as e:
                    raise NaturalKeyTestHelperError(
                        f"{e} See {obj._meta.label_lower}. Got {options}."
                    )

    def nk_test_natural_key_by_schedule(self, visits=None, visit_attr=None):
        """A wrapper method for natural_keys that use
        the enrollment instance to test each CRF in each visit
        in the schedule linked to the enrollment model.
        """
        complete_required_crfs = {}
        for visit in visits:
            complete_required_crfs.update(
                {
                    visit.visit_code: self.complete_required_crfs(
                        visit_code=visit.visit_code,
                        visit=visit,
                        visit_attr=visit_attr,
                        subject_identifier=visit.subject_identifier,
                    )
                }
            )
        self.nk_test_natural_keys(complete_required_crfs)



from typing import Any


class FieldDoesNotExistTestError(Exception):
    pass


def validate_fields_exists_or_raise(cleaned_data: dict, model_cls: Any) -> None:
    if not_found := [
        key
        for key in cleaned_data
        if key not in [fld.name for fld in model_cls._meta.get_fields()]
    ]:
        raise FieldDoesNotExistTestError(f"Invalid field in cleaned data. Got {not_found}")
    return None



import string
from datetime import date, datetime
from random import choice

from dateutil.relativedelta import relativedelta
from django.apps import apps as django_apps
from edc_consent.site_consents import site_consents
from edc_constants.constants import FEMALE, MALE
from edc_utils import get_utcnow
from faker.providers import BaseProvider


class EdcBaseProvider(BaseProvider):
    @property
    def consent_model(self):
        return django_apps.get_app_config("edc_base_test").consent_model

    @staticmethod
    def gender() -> str:
        return choice([FEMALE, MALE])  # nosec B311

    @staticmethod
    def initials() -> str:
        return choice(list(string.ascii_uppercase)) + choice(  # nosec B311
            list(string.ascii_uppercase)
        )

    def dob_for_consenting_adult(self) -> date:
        consent = site_consents.get_consent(
            consent_model=self.consent_model, report_datetime=get_utcnow()
        )
        years = choice(range(consent.age_is_adult, consent.age_max + 1))  # nosec B311
        return (get_utcnow() - relativedelta(years=years)).date()

    def dob_for_consenting_minor(self) -> date:
        consent = site_consents.get_consent(self.consent_model, report_datetime=get_utcnow())
        years = consent.age_min - 1
        return (get_utcnow() - relativedelta(years=years)).date()

    def age_for_consenting_adult(self) -> int:
        consent = site_consents.get_consent(
            consent_model=self.consent_model, report_datetime=get_utcnow()
        )
        return choice(range(consent.age_is_adult, consent.age_max + 1))  # nosec B311

    def age_for_consenting_minor(self) -> int:
        consent = site_consents.get_consent(
            consent_model=self.consent_model, report_datetime=get_utcnow()
        )
        return choice(range(consent.age_min, consent.age_is_adult + 1))  # nosec B311

    @staticmethod
    def yesterday() -> datetime:
        return get_utcnow() - relativedelta(days=1)

    @staticmethod
    def last_week() -> datetime:
        return get_utcnow() - relativedelta(weeks=1)

    @staticmethod
    def last_month() -> datetime:
        return get_utcnow() - relativedelta(months=1)

    @staticmethod
    def two_months_ago() -> datetime:
        return get_utcnow() - relativedelta(months=2)

    @staticmethod
    def three_months_ago() -> datetime:
        return get_utcnow() - relativedelta(months=3)

    @staticmethod
    def six_months_ago() -> datetime:
        return get_utcnow() - relativedelta(months=6)

    @staticmethod
    def twelve_months_ago() -> datetime:
        return get_utcnow() - relativedelta(months=12)



from dateutil.relativedelta import relativedelta
from django.apps import apps as django_apps
from django.conf import settings
from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist
from edc_appointment.models import Appointment
from edc_utils import get_utcnow
from edc_visit_schedule.constants import DAY1
from edc_visit_tracking.constants import SCHEDULED, UNSCHEDULED


class SubjectVisitTestCaseMixin:
    @property
    def subject_visit_model_cls(self):
        return django_apps.get_model(settings.SUBJECT_VISIT_MODEL)

    @property
    def subject_screening_model_cls(self):
        return django_apps.get_model(settings.SUBJECT_SCREENING_MODEL)

    @property
    def subject_consent_model_cls(self):
        return django_apps.get_model(settings.SUBJECT_CONSENT_MODEL)

    def get_subject_screening(
        self, screening_identifier=None, age_in_years=None, initials=None, **kwargs
    ):
        age_in_years = age_in_years or 25
        initials = initials or "XX"
        try:
            subject_screening = self.subject_screening_model_cls.objects.get(
                screening_identifier=screening_identifier
            )
        except ObjectDoesNotExist:
            subject_screening = self.subject_screening_model_cls.objects.create(
                screening_identifier=screening_identifier,
                age_in_years=age_in_years,
                initials=initials,
                **kwargs,
            )
        except LookupError:
            subject_screening = None
        return subject_screening

    def get_subject_consent(self, subject_screening, consent_datetime=None, **kwargs):
        consent_datetime = consent_datetime or subject_screening.report_datetime
        dob = (consent_datetime or get_utcnow()).date() - relativedelta(
            years=subject_screening.age_in_years
        )
        options = dict(
            user_created="erikvw",
            user_modified="erikvw",
            dob=dob,
            site=Site.objects.get(id=settings.SITE_ID),
            consent_datetime=consent_datetime,
            screening_identifier=subject_screening.screening_identifier,
            initials=subject_screening.initials,
        )
        options.update(**kwargs)
        return self.subject_consent_model_cls.objects.create(**options)

    @staticmethod
    def get_appointment(**kwargs):
        return Appointment.objects.get(**kwargs)

    def get_subject_visit(
        self,
        visit_code=None,
        visit_code_sequence=None,
        subject_screening=None,
        subject_consent=None,
        reason=None,
        appointment=None,
        appt_datetime=None,
        report_datetime=None,
    ):
        reason = reason or SCHEDULED
        if not appointment:
            subject_consent = subject_consent or self.get_subject_consent(
                subject_screening or self.get_subject_screening()
            )
            options = dict(
                subject_identifier=subject_consent.subject_identifier,
                visit_code=visit_code or DAY1,
                visit_code_sequence=(
                    visit_code_sequence if visit_code_sequence is not None else 0
                ),
                reason=reason,
            )
            if appt_datetime:
                options.update(appt_datetime=appt_datetime)
            appointment = self.get_appointment(**options)
        try:
            return self.subject_visit_model_cls.objects.get(appointment=appointment)
        except ObjectDoesNotExist:
            return self.subject_visit_model_cls.objects.create(
                appointment=appointment,
                reason=reason,
                report_datetime=report_datetime or appt_datetime or appointment.appt_datetime,
            )

    def get_next_subject_visit(
        self,
        subject_visit=None,
        reason=None,
        appt_datetime=None,
        report_datetime=None,
    ):
        visit_code = (
            subject_visit.appointment.visit_code
            if reason == UNSCHEDULED
            else subject_visit.appointment.next.visit_code
        )
        # visit_code_sequence will increment in get_subject_visit
        visit_code_sequence = (
            subject_visit.appointment.visit_code_sequence if reason == UNSCHEDULED else 0
        )
        return self.get_subject_visit(
            visit_code=visit_code,
            visit_code_sequence=visit_code_sequence,
            reason=reason,
            appt_datetime=appt_datetime,
            subject_screening=self.subject_screening_model_cls.objects.get(
                subject_identifier=subject_visit.subject_identifier
            ),
            subject_consent=self.subject_consent_model_cls.objects.get(
                subject_identifier=subject_visit.subject_identifier
            ),
            report_datetime=report_datetime,
        )
